searchNodes=[{"doc":"Goal is a parameter validation library based on Ecto. Goal takes the params (e.g. from an Phoenix controller), validates them against a schema, and returns an atom-based map or an error changeset. It's based on Ecto , so every validation that you have for database fields can be applied in validating parameters. Goal is different from other validation libraries because of its syntax, being Ecto-based, and validating data using functions from Ecto.Changeset instead of building embedded Ecto.Schema s in the background. Additionally, Goal allows you to configure your own regexes. This is helpful in case of backward compatibility, where Goal's defaults might not match your production system's behavior. Usage Goal's entry point is Goal.validate_params/2 , which receives the parameters and a validation schema. The parameters must be a map, and can be string-based or atom-based. Goal needs a validation schema (also a map) to parse and validate the parameters. You can build one with the defschema macro: defmodule MyApp.SomeController do import Goal import Goal.Syntax def create ( conn , params ) do with { :ok , attrs } &lt;- validate_params ( params , schema ( ) ) do ... end end defp schema do defschema do required :uuid , :string , format : :uuid required :name , :string , min : 3 , max : 3 optional :age , :integer , min : 0 , max : 120 optional :gender , :enum , values : [ &quot;female&quot; , &quot;male&quot; , &quot;non-binary&quot; ] optional :data , :map do required :color , :string optional :money , :decimal optional :height , :float end end end end The defschema macro converts the given structure into a validation schema at compile-time. You can also use the basic syntax like in the example below. The basic syntax is what defschema compiles to. defmodule MyApp.SomeController do import Goal @schema %{ id : [ format : :uuid , required : true ] , name : [ min : 3 , max : 20 , required : true ] , age : [ type : :integer , min : 0 , max : 120 ] , gender : [ type : :enum , values : [ &quot;female&quot; , &quot;male&quot; , &quot;non-binary&quot; ] ] , data : [ type : :map , properties : %{ color : [ required : true ] , money : [ type : :decimal ] , height : [ type : :float ] } ] } def create ( conn , params ) do with { :ok , attrs } &lt;- validate_params ( params , @schema ) do ... end end end Features Bring your own regex Goal has sensible defaults for string format validation. If you'd like to use your own regex, e.g. for validating email addresses or passwords, then you can add your own regex in the configuration: config :goal , uuid_regex : ~r/^[[:alpha:]]+$/ , email_regex : ~r/^[[:alpha:]]+$/ , password_regex : ~r/^[[:alpha:]]+$/ , url_regex : ~r/^[[:alpha:]]+$/ Deeply nested maps Goal efficiently builds error changesets for nested maps, and has support for lists of nested maps. There is no limitation on depth. params = %{ &quot;nested_map&quot; =&gt; %{ &quot;map&quot; =&gt; %{ &quot;inner_map&quot; =&gt; %{ &quot;id&quot; =&gt; 123 , &quot;list&quot; =&gt; [ 1 , 2 , 3 ] } } } } schema = %{ nested_map : [ type : :map , properties : %{ inner_map : [ type : :map , properties : %{ map : [ type : :map , properties : %{ id : [ type : :integer , required : true ] , list : [ type : { :array , :integer } ] } ] } ] } ] } iex(3)&gt; Goal . validate_params ( params , schema ) { :ok , %{ nested_map : %{ inner_map : %{ map : %{ id : 123 , list : [ 1 , 2 , 3 ] } } } } } Use defschema to reduce boilerplate Goal provides a macro called Goal.Syntax.defschema/1 to build validation schemas without all the boilerplate code. The previous example of deeply nested maps can be rewritten to: import Goal.Syntax params = %{ ... } schema = defschema do optional :nested_map , :map do optional :inner_map , :map do optional :map , :map do required :id , :integer optional :list , { :array , :integer } end end end end iex(3)&gt; Goal . validate_params ( params , schema ) { :ok , %{ nested_map : %{ inner_map : %{ map : %{ id : 123 , list : [ 1 , 2 , 3 ] } } } } } Readable error messages Use Goal.traverse_errors/2 to build readable errors. Phoenix by default uses Ecto.Changeset.traverse_errors/2 , which works for embedded Ecto schemas but not for the plain nested maps used by Goal. Goal's traverse_errors/2 is compatible with (embedded) Ecto.Schema s, so you don't have to make any changes to your existing logic. def translate_errors ( changeset ) do Goal . traverse_errors ( changeset , &amp; translate_error / 1 ) end Available validations The field types and available validations are: Field type Validations Description :string :equals string value :is string length :min minimum string length :max maximum string length :trim oolean to remove leading and trailing spaces :squish boolean to trim and collapse spaces :format :uuid , :email , :password , :url :subset list of required strings :included list of allowed strings :excluded list of disallowed strings :integer :equals integer value :is integer value :min minimum integer value :max maximum integer value :greater_than minimum integer value :less_than maximum integer value :greater_than_or_equal_to minimum integer value :less_than_or_equal_to maximum integer value :equal_to integer value :not_equal_to integer value :subset list of required integers :included list of allowed integers :excluded list of disallowed integers :float all of the integer validations :decimal all of the integer validations :boolean :equals boolean value :date :equals date value :time :equals time value :enum :values list of allowed values :map :properties use :properties to define the fields {:array, :map} :properties use :properties to define the fields {:array, inner_type} inner_type can be any of the basic types More basic types See Ecto.Schema for the full list The default basic type is :string . You don't have to define this field if you are using the basic syntax. All field types, exluding :map and {:array, :map} , can use :equals , :subset , :included , :excluded validations. Credits This library is based on Ecto and I had to copy and adapt Ecto.Changeset.traverse_errors/2 . Thanks for making such an awesome library! 🙇","ref":"Goal.html","title":"Goal","type":"module"},{"doc":"Traverses changeset errors and applies the given function to error messages. Examples iex&gt; traverse_errors ( changeset , fn { msg , opts } -&gt; ...&gt; Regex . replace ( ~r&quot;%{( \\w +)}&quot; , msg , fn _ , key -&gt; ...&gt; opts |&gt; Keyword . get ( String . to_existing_atom ( key ) , key ) |&gt; to_string ( ) ...&gt; end ) ...&gt; end ) %{ title : [ &quot;should be at least 3 characters&quot; ] }","ref":"Goal.html#traverse_errors/2","title":"Goal.traverse_errors/2","type":"function"},{"doc":"Validates parameters against a schema. Examples iex&gt; validate_params ( %{ &quot;email&quot; =&gt; &quot;jane@example.com&quot; } , %{ email : [ format : :email ] } ) { :ok , %{ email : &quot;jane@example.com&quot; } } iex&gt; validate_params ( %{ &quot;email&quot; =&gt; &quot;invalid&quot; } , %{ email : [ format : :email ] } ) { :error , % Ecto.Changeset { valid? : false , errors : [ email : { &quot;has invalid format&quot; , ... } ] } }","ref":"Goal.html#validate_params/2","title":"Goal.validate_params/2","type":"function"},{"doc":"","ref":"Goal.html#t:error/0","title":"Goal.error/0","type":"type"},{"doc":"","ref":"Goal.html#t:params/0","title":"Goal.params/0","type":"type"},{"doc":"","ref":"Goal.html#t:schema/0","title":"Goal.schema/0","type":"type"},{"doc":"Goal.Changeset contains an adapted version of Ecto.Changeset.traverse_errors/2 . You can use traverse_errors/2 to build all errors into a series of nested maps.","ref":"Goal.Changeset.html","title":"Goal.Changeset","type":"module"},{"doc":"Traverses changeset errors and applies the given function to error messages. This function is particularly useful when associations, embeds, maps and nested maps are cast in the changeset as it will traverse all associations, embeds, maps and nested maps and place all errors in a series of nested maps. A changeset is supplied along with a function to apply to each error message as the changeset is traversed. The error message function receives an error tuple {msg, opts} , for example: { &quot;should be at least %{count} characters&quot; , [ count : 3 , validation : :length , min : 3 ] } Examples iex&gt; traverse_errors ( changeset , fn { msg , opts } -&gt; ...&gt; Regex . replace ( ~r&quot;%{( \\w +)}&quot; , msg , fn _ , key -&gt; ...&gt; opts |&gt; Keyword . get ( String . to_existing_atom ( key ) , key ) |&gt; to_string ( ) ...&gt; end ) ...&gt; end ) %{ title : [ &quot;should be at least 3 characters&quot; ] } Optionally function can accept three arguments: changeset , field and error tuple {msg, opts} . It is useful whenever you want to extract validations rules from changeset.validations to build detailed error description. This function and documentation is gratefully copied and adapted from Ecto.Changeset.traverse_errors/2 🙇","ref":"Goal.Changeset.html#traverse_errors/2","title":"Goal.Changeset.traverse_errors/2","type":"function"},{"doc":"","ref":"Goal.Changeset.html#t:action/0","title":"Goal.Changeset.action/0","type":"type"},{"doc":"","ref":"Goal.Changeset.html#t:constraint/0","title":"Goal.Changeset.constraint/0","type":"type"},{"doc":"","ref":"Goal.Changeset.html#t:data/0","title":"Goal.Changeset.data/0","type":"type"},{"doc":"","ref":"Goal.Changeset.html#t:error/0","title":"Goal.Changeset.error/0","type":"type"},{"doc":"","ref":"Goal.Changeset.html#t:t/0","title":"Goal.Changeset.t/0","type":"type"},{"doc":"","ref":"Goal.Changeset.html#t:t/1","title":"Goal.Changeset.t/1","type":"type"},{"doc":"","ref":"Goal.Changeset.html#t:types/0","title":"Goal.Changeset.types/0","type":"type"},{"doc":"Defines regexes.","ref":"Goal.Regex.html","title":"Goal.Regex","type":"module"},{"doc":"Defines string operations.","ref":"Goal.String.html","title":"Goal.String","type":"module"},{"doc":"Returns a string where all leading, trailing, and multiple inner Unicode whitespaces have been removed. Examples iex&gt; squish ( &quot; banana man &quot; ) &quot;banana man&quot;","ref":"Goal.String.html#squish/1","title":"Goal.String.squish/1","type":"function"},{"doc":"Goal.Syntax provides the defschema macro for defining validation schemas. Usage import Goal.Syntax","ref":"Goal.Syntax.html","title":"Goal.Syntax","type":"module"},{"doc":"A macro for defining validation schemas that are generated at compile-time. import Goal.Syntax defp schema do defschema do required :id , :string , format : :uuid required :name , :string optional :age , :integer , min : 0 , max : 120 optional :gender , :enum , values : [ &quot;female&quot; , &quot;male&quot; , &quot;non-binary&quot; ] required :data , :map do required :city , :string optional :birthday , :date end end end","ref":"Goal.Syntax.html#defschema/1","title":"Goal.Syntax.defschema/1","type":"macro"},{"doc":"A library for parsing and validating parameters. Goal takes the params (e.g. from an Phoenix controller), validates them against a schema, and returns an atom-based map or an error changeset. It's based on Ecto , so every validation that you have for database fields can be applied in validating parameters. Goal is different from other validation libraries because of its syntax, being Ecto-based, and validating data using functions from Ecto.Changeset instead of building embedded Ecto.Schema s in the background. Additionally, Goal allows you to configure your own regexes. This is helpful in case of backward compatibility, where Goal's defaults might not match your production system's behavior.","ref":"readme.html","title":"Goal ⚽","type":"extras"},{"doc":"Add goal to the list of dependencies in mix.exs : def deps do [ { :goal , &quot;~&gt; 0.1.1&quot; } ] end","ref":"readme.html#installation","title":"Goal ⚽ - Installation","type":"extras"},{"doc":"Goal's entry point is Goal.validate_params/2 , which receives the parameters and a validation schema. The parameters must be a map, and can be string-based or atom-based. Goal needs a validation schema (also a map) to parse and validate the parameters. You can build one with the defschema macro: defmodule MyApp.SomeController do import Goal import Goal.Syntax def create ( conn , params ) do with { :ok , attrs } &lt;- validate_params ( params , schema ( ) ) do ... end end defp schema do defschema do required :uuid , :string , format : :uuid required :name , :string , min : 3 , max : 3 optional :age , :integer , min : 0 , max : 120 optional :gender , :enum , values : [ &quot;female&quot; , &quot;male&quot; , &quot;non-binary&quot; ] optional :data , :map do required :color , :string optional :money , :decimal optional :height , :float end end end end The defschema macro converts the given structure into a validation schema at compile-time. You can also use the basic syntax like in the example below. The basic syntax is what defschema compiles to. defmodule MyApp.SomeController do import Goal @schema %{ id : [ format : :uuid , required : true ] , name : [ min : 3 , max : 20 , required : true ] , age : [ type : :integer , min : 0 , max : 120 ] , gender : [ type : :enum , values : [ &quot;female&quot; , &quot;male&quot; , &quot;non-binary&quot; ] ] , data : [ type : :map , properties : %{ color : [ required : true ] , money : [ type : :decimal ] , height : [ type : :float ] } ] } def create ( conn , params ) do with { :ok , attrs } &lt;- validate_params ( params , @schema ) do ... end end end","ref":"readme.html#usage","title":"Goal ⚽ - Usage","type":"extras"},{"doc":"Bring your own regex Goal has sensible defaults for string format validation. If you'd like to use your own regex, e.g. for validating email addresses or passwords, then you can add your own regex in the configuration: config :goal , uuid_regex : ~r/^[[:alpha:]]+$/ , email_regex : ~r/^[[:alpha:]]+$/ , password_regex : ~r/^[[:alpha:]]+$/ , url_regex : ~r/^[[:alpha:]]+$/ Deeply nested maps Goal efficiently builds error changesets for nested maps, and has support for lists of nested maps. There is no limitation on depth. params = %{ &quot;nested_map&quot; =&gt; %{ &quot;map&quot; =&gt; %{ &quot;inner_map&quot; =&gt; %{ &quot;id&quot; =&gt; 123 , &quot;list&quot; =&gt; [ 1 , 2 , 3 ] } } } } schema = %{ nested_map : [ type : :map , properties : %{ inner_map : [ type : :map , properties : %{ map : [ type : :map , properties : %{ id : [ type : :integer , required : true ] , list : [ type : { :array , :integer } ] } ] } ] } ] } iex(3)&gt; Goal . validate_params ( params , schema ) { :ok , %{ nested_map : %{ inner_map : %{ map : %{ id : 123 , list : [ 1 , 2 , 3 ] } } } } } Use defschema to reduce boilerplate Goal provides a macro called Goal.Syntax.defschema/1 to build validation schemas without all the boilerplate code. The previous example of deeply nested maps can be rewritten to: import Goal.Syntax params = %{ ... } schema = defschema do optional :nested_map , :map do optional :inner_map , :map do optional :map , :map do required :id , :integer optional :list , { :array , :integer } end end end end iex(3)&gt; Goal . validate_params ( params , schema ) { :ok , %{ nested_map : %{ inner_map : %{ map : %{ id : 123 , list : [ 1 , 2 , 3 ] } } } } } Readable error messages Use Goal.traverse_errors/2 to build readable errors. Phoenix by default uses Ecto.Changeset.traverse_errors/2 , which works for embedded Ecto schemas but not for the plain nested maps used by Goal. Goal's traverse_errors/2 is compatible with (embedded) Ecto.Schema s, so you don't have to make any changes to your existing logic. def translate_errors ( changeset ) do Goal . traverse_errors ( changeset , &amp; translate_error / 1 ) end Available validations The field types and available validations are: Field type Validations Description :string :equals string value :is string length :min minimum string length :max maximum string length :trim oolean to remove leading and trailing spaces :squish boolean to trim and collapse spaces :format :uuid , :email , :password , :url :subset list of required strings :included list of allowed strings :excluded list of disallowed strings :integer :equals integer value :is integer value :min minimum integer value :max maximum integer value :greater_than minimum integer value :less_than maximum integer value :greater_than_or_equal_to minimum integer value :less_than_or_equal_to maximum integer value :equal_to integer value :not_equal_to integer value :subset list of required integers :included list of allowed integers :excluded list of disallowed integers :float all of the integer validations :decimal all of the integer validations :boolean :equals boolean value :date :equals date value :time :equals time value :enum :values list of allowed values :map :properties use :properties to define the fields {:array, :map} :properties use :properties to define the fields {:array, inner_type} inner_type can be any of the basic types More basic types See Ecto.Schema for the full list The default basic type is :string . You don't have to define this field if you are using the basic syntax. All field types, exluding :map and {:array, :map} , can use :equals , :subset , :included , :excluded validations.","ref":"readme.html#features","title":"Goal ⚽ - Features","type":"extras"},{"doc":"[x] Bring your own regex [x] ExDoc documentation [x] Basic syntax optimizations [x] Macro for generating schemas without boilerplate [x] Release v0.1.0 on Hex.pm [ ] Convert incoming params from camelCase to snake_case","ref":"readme.html#roadmap","title":"Goal ⚽ - Roadmap","type":"extras"},{"doc":"This library is based on Ecto and I had to copy and adapt Ecto.Changeset.traverse_errors/2 . Thanks for making such an awesome library! 🙇","ref":"readme.html#credits","title":"Goal ⚽ - Credits","type":"extras"}]